import _Promise from 'babel-runtime/core-js/promise';
import _JSON$stringify from 'babel-runtime/core-js/json/stringify';
import _regeneratorRuntime from 'babel-runtime/regenerator';
import _asyncToGenerator from 'babel-runtime/helpers/asyncToGenerator';
import _classCallCheck from 'babel-runtime/helpers/classCallCheck';
import _createClass from 'babel-runtime/helpers/createClass';
import debugStart from 'debug';
import validator from 'xsd-schema-validator';
import bluebird from 'bluebird';
import redis from 'redis';
import kue from 'kue-scheduler';
import Redlock from 'redlock';

//  .d888
// d88P'
// 888
// 888888.d8888b 88888b.d88b.        .d8888b .d88b. 888d888 .d88b.
// 888   88K     888 '888 '88b      d88P'   d88''88b888P'  d8P  Y8b
// 888   'Y8888b.888  888  888888888888     888  888888    88888888
// 888        X88888  888  888      Y88b.   Y88..88P888    Y8b.
// 888    88888P'888  888  888       'Y8888P 'Y88P' 888     'Y8888

bluebird.promisifyAll(redis.RedisClient.prototype);
bluebird.promisifyAll(redis.Multi.prototype);

var debug = debugStart('core');
/**
 * The fsm-core module provides facilities to persist SCXML models and all its related data in a Redis database
 */

var Core = function () {
    function Core() {
        _classCallCheck(this, Core);
    }

    _createClass(Core, [{
        key: 'init',


        /**
         * Initializes a redis client
         * @method init
         * @memberOf Core
         */
        value: function init(redisOptions, kueOptions) {

            redisOptions.retry_strategy = function (options) {
                if (options.error && options.error.code === 'ECONNREFUSED') {
                    // End reconnecting on a specific error and flush all commands with
                    // a individual error
                    return new Error('The server refused the connection');
                }
                // reconnect after
                return Math.min(options.attempt * 100, 3000);
            };

            this.client = redis.createClient(redisOptions);
            this.queue = kue.createQueue({
                prefix: 'q',
                redis: kueOptions
            });

            this.queue.on('error', function (err) {
                debug('Oops... ', err);
            });

            this.redlock = new Redlock(
            // you should have one client for each redis node
            // in your cluster
            [this.client], {
                // the expected clock drift; for more details
                // see http://redis.io/topics/distlock
                driftFactor: 0.01, // time in ms

                // the max number of times Redlock will attempt
                // to lock a resource before erroring
                retryCount: 200,

                // the time in ms between attempts
                retryDelay: 200, // time in ms

                // the max time in ms randomly added to retries
                // to improve performance under high contention
                // see https://www.awsarchitectureblog.com/2015/03/backoff.html
                retryJitter: 200 // time in ms
            });
        }
    }, {
        key: 'disconnect',
        value: function disconnect() {
            this.client.quit();
            // this.queue.shutdown();
        }
    }, {
        key: '_verifyMachine',
        value: function () {
            var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(machineName) {
                var machineExists;
                return _regeneratorRuntime.wrap(function _callee$(_context) {
                    while (1) {
                        switch (_context.prev = _context.next) {
                            case 0:
                                _context.next = 2;
                                return this.client.sismemberAsync('machineList', machineName);

                            case 2:
                                machineExists = _context.sent;

                                if (machineExists) {
                                    _context.next = 5;
                                    break;
                                }

                                throw new Error('Machine does not exists');

                            case 5:
                            case 'end':
                                return _context.stop();
                        }
                    }
                }, _callee, this);
            }));

            function _verifyMachine(_x) {
                return _ref.apply(this, arguments);
            }

            return _verifyMachine;
        }()
    }, {
        key: '_verifyVersion',
        value: function () {
            var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(machineName, versionKey) {
                var versionExists;
                return _regeneratorRuntime.wrap(function _callee2$(_context2) {
                    while (1) {
                        switch (_context2.prev = _context2.next) {
                            case 0:
                                _context2.next = 2;
                                return this._verifyMachine(machineName);

                            case 2:
                                _context2.next = 4;
                                return this.client.zrankAsync('machine:' + machineName + ':versionList', versionKey);

                            case 4:
                                versionExists = _context2.sent;

                                if (!(versionExists === null)) {
                                    _context2.next = 7;
                                    break;
                                }

                                throw new Error('Version does not exists');

                            case 7:
                            case 'end':
                                return _context2.stop();
                        }
                    }
                }, _callee2, this);
            }));

            function _verifyVersion(_x2, _x3) {
                return _ref2.apply(this, arguments);
            }

            return _verifyVersion;
        }()
    }, {
        key: '_verifyInstance',
        value: function () {
            var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(machineName, versionKey, instanceKey) {
                var instanceExists;
                return _regeneratorRuntime.wrap(function _callee3$(_context3) {
                    while (1) {
                        switch (_context3.prev = _context3.next) {
                            case 0:
                                _context3.next = 2;
                                return this._verifyVersion(machineName, versionKey);

                            case 2:
                                _context3.next = 4;
                                return this.client.zrankAsync('machine:' + machineName + ':version:' + versionKey + ':instanceList', instanceKey);

                            case 4:
                                instanceExists = _context3.sent;

                                if (!(instanceExists === null)) {
                                    _context3.next = 7;
                                    break;
                                }

                                throw new Error('Instance does not exists');

                            case 7:
                            case 'end':
                                return _context3.stop();
                        }
                    }
                }, _callee3, this);
            }));

            function _verifyInstance(_x4, _x5, _x6) {
                return _ref3.apply(this, arguments);
            }

            return _verifyInstance;
        }()
    }, {
        key: '_verifySnapshot',
        value: function () {
            var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee4(machineName, versionKey, instanceKey, snapshotKey) {
                var snapshotExists;
                return _regeneratorRuntime.wrap(function _callee4$(_context4) {
                    while (1) {
                        switch (_context4.prev = _context4.next) {
                            case 0:
                                _context4.next = 2;
                                return this._verifyInstance(machineName, versionKey, instanceKey);

                            case 2:
                                _context4.next = 4;
                                return this.client.zrankAsync('machine:' + machineName + ':version:' + versionKey + ':instance:' + instanceKey + ':snapshotList', snapshotKey);

                            case 4:
                                snapshotExists = _context4.sent;

                                if (!(snapshotExists === null)) {
                                    _context4.next = 7;
                                    break;
                                }

                                throw new Error('Snapshot does not exists');

                            case 7:
                            case 'end':
                                return _context4.stop();
                        }
                    }
                }, _callee4, this);
            }));

            function _verifySnapshot(_x7, _x8, _x9, _x10) {
                return _ref4.apply(this, arguments);
            }

            return _verifySnapshot;
        }()
        /**
         * Retrieve the repository manifest.json file as a JavasScript Object
         * @method getManifest
         * @memberOf Core
         * @returns {Object} The manifest Object
         */

    }, {
        key: 'getManifest',
        value: function () {
            var _ref5 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee5() {
                return _regeneratorRuntime.wrap(function _callee5$(_context5) {
                    while (1) {
                        switch (_context5.prev = _context5.next) {
                            case 0:
                                _context5.next = 2;
                                return this.client.hgetallAsync("manifest");

                            case 2:
                                return _context5.abrupt('return', _context5.sent);

                            case 3:
                            case 'end':
                                return _context5.stop();
                        }
                    }
                }, _callee5, this);
            }));

            function getManifest() {
                return _ref5.apply(this, arguments);
            }

            return getManifest;
        }()

        /**
         * Get the names of all of the machines in the repository
         * @method getMachinesNames
         * @memberOf Core
         * @returns {Promise} An array with all the machine's names
         */

    }, {
        key: 'getMachinesNames',
        value: function () {
            var _ref6 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee6() {
                return _regeneratorRuntime.wrap(function _callee6$(_context6) {
                    while (1) {
                        switch (_context6.prev = _context6.next) {
                            case 0:
                                _context6.next = 2;
                                return this.client.smembersAsync('machineList');

                            case 2:
                                return _context6.abrupt('return', _context6.sent);

                            case 3:
                            case 'end':
                                return _context6.stop();
                        }
                    }
                }, _callee6, this);
            }));

            function getMachinesNames() {
                return _ref6.apply(this, arguments);
            }

            return getMachinesNames;
        }()

        /**
         * Add a new machine to the repository
         * @method addMachine
         * @memberOf Core
         * @param {String} machineName The name of the new machine
         * @returns {Promise}
         */

    }, {
        key: 'addMachine',
        value: function () {
            var _ref7 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee7(machineName) {
                var machineExists, baseSCXML, info, lock;
                return _regeneratorRuntime.wrap(function _callee7$(_context7) {
                    while (1) {
                        switch (_context7.prev = _context7.next) {
                            case 0:
                                _context7.next = 2;
                                return this.client.sismemberAsync('machineList', machineName);

                            case 2:
                                machineExists = _context7.sent;

                                if (!machineExists) {
                                    _context7.next = 5;
                                    break;
                                }

                                throw new Error('Machine already exists');

                            case 5:
                                _context7.next = 7;
                                return this.client.getAsync('baseSCXML');

                            case 7:
                                baseSCXML = _context7.sent;
                                info = {
                                    isSealed: false
                                };

                                //lock

                                _context7.next = 11;
                                return this.redlock.lock('lock:machineList', 100);

                            case 11:
                                lock = _context7.sent;
                                _context7.next = 14;
                                return this.client.hmsetAsync('machine:' + machineName + ':version:version1', 'model', baseSCXML, 'info', _JSON$stringify(info));

                            case 14:
                                _context7.next = 16;
                                return this.client.sadd('machineList', machineName);

                            case 16:
                                _context7.next = 18;
                                return this.client.zaddAsync('machine:' + machineName + ':versionList', 1, 'version1');

                            case 18:
                                _context7.next = 20;
                                return lock.unlock();

                            case 20:
                                _context7.next = 22;
                                return this.client.hincrbyAsync('manifest', 'machineCount', 1);

                            case 22:
                                _context7.next = 24;
                                return this.client.hincrbyAsync('manifest', 'versionCount', 1);

                            case 24:
                            case 'end':
                                return _context7.stop();
                        }
                    }
                }, _callee7, this);
            }));

            function addMachine(_x11) {
                return _ref7.apply(this, arguments);
            }

            return addMachine;
        }()

        /**
         * Get the keys of all of the versions of machine in the repository
         * @method getVersionsKeys
         * @memberOf Core
         * @param {String} machineName The name of the machine to get the version's keys
         * @returns {Array} An array with all the version's keys of the machine
         */

    }, {
        key: 'getVersionsKeys',
        value: function () {
            var _ref8 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee8(machineName) {
                return _regeneratorRuntime.wrap(function _callee8$(_context8) {
                    while (1) {
                        switch (_context8.prev = _context8.next) {
                            case 0:
                                _context8.next = 2;
                                return this._verifyMachine(machineName);

                            case 2:
                                _context8.next = 4;
                                return this.client.zrangeAsync('machine:' + machineName + ':versionList', 0, -1);

                            case 4:
                                return _context8.abrupt('return', _context8.sent);

                            case 5:
                            case 'end':
                                return _context8.stop();
                        }
                    }
                }, _callee8, this);
            }));

            function getVersionsKeys(_x12) {
                return _ref8.apply(this, arguments);
            }

            return getVersionsKeys;
        }()
        /**
         * Retrieve the version info.json file as a JavasScript Object
         * @method getVersionInfo
         * @memberOf Core
         * @param {String} machineName The name of the machine
         * @param {String} versionKey The key of the version
         * @returns {Object} The info Object
         */

    }, {
        key: 'getVersionInfo',
        value: function () {
            var _ref9 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee9(machineName, versionKey) {
                var info;
                return _regeneratorRuntime.wrap(function _callee9$(_context9) {
                    while (1) {
                        switch (_context9.prev = _context9.next) {
                            case 0:
                                _context9.next = 2;
                                return this._verifyVersion(machineName, versionKey);

                            case 2:
                                _context9.next = 4;
                                return this.client.hgetAsync('machine:' + machineName + ':version:' + versionKey, 'info');

                            case 4:
                                info = _context9.sent;
                                return _context9.abrupt('return', JSON.parse(info));

                            case 6:
                            case 'end':
                                return _context9.stop();
                        }
                    }
                }, _callee9, this);
            }));

            function getVersionInfo(_x13, _x14) {
                return _ref9.apply(this, arguments);
            }

            return getVersionInfo;
        }()

        /**
         * Update the version info.json file using a JavasScript Object
         * @method setVersionInfo
         * @memberOf Core
         * @param {String} machineName The name of the machine
         * @param {String} versionKey The key of the version
         * @param {Object} info The info Object to save
         * @param {String} message If supplied it is used as the message for the commit
         * @returns {Promise}
         */

    }, {
        key: 'setVersionInfo',
        value: function () {
            var _ref10 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee10(machineName, versionKey, info) {
                return _regeneratorRuntime.wrap(function _callee10$(_context10) {
                    while (1) {
                        switch (_context10.prev = _context10.next) {
                            case 0:
                                _context10.next = 2;
                                return this._verifyVersion(machineName, versionKey);

                            case 2:
                                _context10.next = 4;
                                return this.client.hsetAsync('machine:' + machineName + ':version:' + versionKey, 'info', _JSON$stringify(info));

                            case 4:
                            case 'end':
                                return _context10.stop();
                        }
                    }
                }, _callee10, this);
            }));

            function setVersionInfo(_x15, _x16, _x17) {
                return _ref10.apply(this, arguments);
            }

            return setVersionInfo;
        }()

        /**
         * Add a new version to a machine
         * @method addVersion
         * @memberOf Core
         * @param {String} machineName The name of the machine
         * @returns {Promise} The version key
         */

    }, {
        key: 'addVersion',
        value: function () {
            var _ref11 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee11(machineName) {
                var baseSCXML, info, lock, count, previousVersionInfo, newVersionNumber, newVersionName;
                return _regeneratorRuntime.wrap(function _callee11$(_context11) {
                    while (1) {
                        switch (_context11.prev = _context11.next) {
                            case 0:
                                _context11.next = 2;
                                return this._verifyMachine(machineName);

                            case 2:
                                _context11.next = 4;
                                return this.client.getAsync('baseSCXML');

                            case 4:
                                baseSCXML = _context11.sent;
                                info = {
                                    isSealed: false
                                };

                                //lock

                                _context11.next = 8;
                                return this.redlock.lock('lock:machine:' + machineName + ':versionList', 100);

                            case 8:
                                lock = _context11.sent;
                                _context11.next = 11;
                                return this.client.zcountAsync('machine:' + machineName + ':versionList', '-inf', '+inf');

                            case 11:
                                count = _context11.sent;
                                _context11.next = 14;
                                return this.getVersionInfo(machineName, 'version' + count);

                            case 14:
                                previousVersionInfo = _context11.sent;

                                if (previousVersionInfo.isSealed) {
                                    _context11.next = 17;
                                    break;
                                }

                                throw new Error('The last versions is not sealed yet');

                            case 17:
                                newVersionNumber = count + 1;
                                newVersionName = 'version' + newVersionNumber;
                                _context11.next = 21;
                                return this.client.hmsetAsync('machine:' + machineName + ':version:' + newVersionName, 'model', baseSCXML, 'info', _JSON$stringify(info));

                            case 21:
                                _context11.next = 23;
                                return this.client.zaddAsync('machine:' + machineName + ':versionList', newVersionNumber, newVersionName);

                            case 23:
                                _context11.next = 25;
                                return lock.unlock();

                            case 25:
                                _context11.next = 27;
                                return this.client.hincrbyAsync('manifest', 'versionCount', 1);

                            case 27:
                                return _context11.abrupt('return', 'version' + newVersionNumber);

                            case 28:
                            case 'end':
                                return _context11.stop();
                        }
                    }
                }, _callee11, this);
            }));

            function addVersion(_x18) {
                return _ref11.apply(this, arguments);
            }

            return addVersion;
        }()

        /**
         * Seal a version of a machine
         * @method addMachine
         * @memberOf Core
         * @param {String} machineName The name of the machine
         * @param {String} versionKey The key of the version
         * @returns {Promise}
         */

    }, {
        key: 'sealVersion',
        value: function () {
            var _ref12 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee12(machineName, versionKey) {
                var info, model, isValid;
                return _regeneratorRuntime.wrap(function _callee12$(_context12) {
                    while (1) {
                        switch (_context12.prev = _context12.next) {
                            case 0:
                                _context12.next = 2;
                                return this._verifyVersion(machineName, versionKey);

                            case 2:

                                debug('Attempting to seal the version "%s" of the machine "%s"', versionKey, machineName);
                                _context12.next = 5;
                                return this.getVersionInfo(machineName, versionKey);

                            case 5:
                                info = _context12.sent;

                                if (!info.isSealed) {
                                    _context12.next = 8;
                                    break;
                                }

                                throw new Error('Version is already sealed');

                            case 8:
                                _context12.next = 10;
                                return this.getVersionSCXML(machineName, versionKey);

                            case 10:
                                model = _context12.sent;
                                _context12.next = 13;
                                return this.isSCXMLValid(model);

                            case 13:
                                isValid = _context12.sent;

                                if (isValid) {
                                    _context12.next = 16;
                                    break;
                                }

                                throw new Error('The model is not valid.');

                            case 16:

                                info.isSealed = true;
                                _context12.next = 19;
                                return this.setVersionInfo(machineName, versionKey, info);

                            case 19:
                                debug('The version "%s" of the machine "%s" was sealed successfully', versionKey, machineName);

                            case 20:
                            case 'end':
                                return _context12.stop();
                        }
                    }
                }, _callee12, this);
            }));

            function sealVersion(_x19, _x20) {
                return _ref12.apply(this, arguments);
            }

            return sealVersion;
        }()

        /**
         * Retrieve the version model.scxml file as a String
         * @method getVersionSCXML
         * @memberOf Core
         * @param {String} machineName The name of the machine
         * @param {String} versionKey The key of the version
         * @returns {String} The model
         */

    }, {
        key: 'getVersionSCXML',
        value: function () {
            var _ref13 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee13(machineName, versionKey) {
                return _regeneratorRuntime.wrap(function _callee13$(_context13) {
                    while (1) {
                        switch (_context13.prev = _context13.next) {
                            case 0:
                                _context13.next = 2;
                                return this._verifyVersion(machineName, versionKey);

                            case 2:
                                _context13.next = 4;
                                return this.client.hgetAsync('machine:' + machineName + ':version:' + versionKey, 'model');

                            case 4:
                                return _context13.abrupt('return', _context13.sent);

                            case 5:
                            case 'end':
                                return _context13.stop();
                        }
                    }
                }, _callee13, this);
            }));

            function getVersionSCXML(_x21, _x22) {
                return _ref13.apply(this, arguments);
            }

            return getVersionSCXML;
        }()

        /**
         * Update the version model.scxml file using a String
         * @method setVersionSCXML
         * @memberOf Core
         * @param {String} model The model
         * @param {String} machineName The name of the machine
         * @param {String} versionKey The key of the version
         * @param {String} message If supplied it is used as the message for the commit
         * @returns {Promise}
         */

    }, {
        key: 'setVersionSCXML',
        value: function () {
            var _ref14 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee14(machineName, versionKey, model) {
                var info;
                return _regeneratorRuntime.wrap(function _callee14$(_context14) {
                    while (1) {
                        switch (_context14.prev = _context14.next) {
                            case 0:
                                _context14.next = 2;
                                return this._verifyVersion(machineName, versionKey);

                            case 2:
                                _context14.next = 4;
                                return this.getVersionInfo(machineName, versionKey);

                            case 4:
                                info = _context14.sent;

                                if (!info.isSealed) {
                                    _context14.next = 7;
                                    break;
                                }

                                throw new Error('Cannot change the version SCXML because the version is sealed.');

                            case 7:
                                this.client.hsetAsync('machine:' + machineName + ':version:' + versionKey, 'model', model);

                            case 8:
                            case 'end':
                                return _context14.stop();
                        }
                    }
                }, _callee14, this);
            }));

            function setVersionSCXML(_x23, _x24, _x25) {
                return _ref14.apply(this, arguments);
            }

            return setVersionSCXML;
        }()

        /**
         * Validates SCXML markup as a string
         * @method isSCXMLValid
         * @memberOf Core
         * @param {String} model A string with the SCXML document to validate
         * @returns {Promise} True if the SCXML is valid false otherwise
         */

    }, {
        key: 'isSCXMLValid',
        value: function () {
            var _ref15 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee15(model) {
                return _regeneratorRuntime.wrap(function _callee15$(_context15) {
                    while (1) {
                        switch (_context15.prev = _context15.next) {
                            case 0:
                                _context15.next = 2;
                                return new _Promise(function (resolve, reject) {
                                    if (model === '') {
                                        reject('Model is empty');
                                        return;
                                    }

                                    validator.validateXML(model, __dirname + '/xmlSchemas/scxml.xsd', function (err, result) {
                                        if (err) {
                                            reject(err);
                                            return;
                                        }
                                        resolve(result.valid);
                                    });
                                });

                            case 2:
                                return _context15.abrupt('return', _context15.sent);

                            case 3:
                            case 'end':
                                return _context15.stop();
                        }
                    }
                }, _callee15, this);
            }));

            function isSCXMLValid(_x26) {
                return _ref15.apply(this, arguments);
            }

            return isSCXMLValid;
        }()

        /**
         * Gets the keys of all of the instances of a version of the machine in the repository
         * @method getInstancesKeys
         * @memberOf Core
         * @param {String} machineName The name of the machine to get the instances's keys
         * @param {String} versionKey The key of the version to get the instances's keys
         * @returns {Array} An array with all the instance's keys of the the version
         */

    }, {
        key: 'getInstancesKeys',
        value: function () {
            var _ref16 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee16(machineName, versionKey) {
                return _regeneratorRuntime.wrap(function _callee16$(_context16) {
                    while (1) {
                        switch (_context16.prev = _context16.next) {
                            case 0:
                                _context16.next = 2;
                                return this._verifyVersion(machineName, versionKey);

                            case 2:
                                _context16.next = 4;
                                return this.client.zrangeAsync('machine:' + machineName + ':version:' + versionKey + ':instanceList', 0, -1);

                            case 4:
                                return _context16.abrupt('return', _context16.sent);

                            case 5:
                            case 'end':
                                return _context16.stop();
                        }
                    }
                }, _callee16, this);
            }));

            function getInstancesKeys(_x27, _x28) {
                return _ref16.apply(this, arguments);
            }

            return getInstancesKeys;
        }()

        /**
         * Retrieve the instance's info.json file as a JavasScript Object
         * @method getInstanceInfo
         * @memberOf Core
         * @param {String} machineName The name of the machine
         * @param {String} versionKey The key of the version
         * @param {String} instanceKey The key of the instance
         * @returns {Object} The info Object
         */

    }, {
        key: 'getInstanceInfo',
        value: function () {
            var _ref17 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee17(machineName, versionKey, instanceKey) {
                var info;
                return _regeneratorRuntime.wrap(function _callee17$(_context17) {
                    while (1) {
                        switch (_context17.prev = _context17.next) {
                            case 0:
                                _context17.next = 2;
                                return this._verifyInstance(machineName, versionKey, instanceKey);

                            case 2:
                                _context17.next = 4;
                                return this.client.hgetAsync('machine:' + machineName + ':version:' + versionKey + ':instance:' + instanceKey, 'info');

                            case 4:
                                info = _context17.sent;
                                return _context17.abrupt('return', JSON.parse(info));

                            case 6:
                            case 'end':
                                return _context17.stop();
                        }
                    }
                }, _callee17, this);
            }));

            function getInstanceInfo(_x29, _x30, _x31) {
                return _ref17.apply(this, arguments);
            }

            return getInstanceInfo;
        }()

        /**
         * Update the instance's info.json file using a JavasScript Object
         * @method setInstanceInfo
         * @memberOf Core
         * @param {String} machineName The name of the machine
         * @param {String} versionKey The key of the version
         * @param {String} instanceKey The key of the instance
         * @param {Object} info The info Object to save
         * @param {String} message If supplied it is used as the message for the commit
         * @returns {Promise}
         */

    }, {
        key: 'setInstanceInfo',
        value: function () {
            var _ref18 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee18(machineName, versionKey, instanceKey, info) {
                return _regeneratorRuntime.wrap(function _callee18$(_context18) {
                    while (1) {
                        switch (_context18.prev = _context18.next) {
                            case 0:
                                _context18.next = 2;
                                return this._verifyInstance(machineName, versionKey, instanceKey);

                            case 2:
                                _context18.next = 4;
                                return this.client.hsetAsync('machine:' + machineName + ':version:' + versionKey + ':instance:' + instanceKey, 'info', _JSON$stringify(info));

                            case 4:
                            case 'end':
                                return _context18.stop();
                        }
                    }
                }, _callee18, this);
            }));

            function setInstanceInfo(_x32, _x33, _x34, _x35) {
                return _ref18.apply(this, arguments);
            }

            return setInstanceInfo;
        }()

        /**
         * Add a new instance to a version of a machine
         * @method addInstance
         * @memberOf Core
         * @param {String} machineName The name of the machine
         * @param {String} versionKey The key of the version
         * @returns {Promise} The instance key
         */

    }, {
        key: 'addInstance',
        value: function () {
            var _ref19 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee19(machineName, versionKey) {
                var versionInfo, info, lock, count, newInstanceNumber, newInstanceName;
                return _regeneratorRuntime.wrap(function _callee19$(_context19) {
                    while (1) {
                        switch (_context19.prev = _context19.next) {
                            case 0:

                                debug('Adding a new instance to the ' + versionKey + ' of the "' + machineName + '" machine');
                                _context19.next = 3;
                                return this._verifyVersion(machineName, versionKey);

                            case 3:
                                _context19.next = 5;
                                return this.getVersionInfo(machineName, versionKey);

                            case 5:
                                versionInfo = _context19.sent;

                                if (versionInfo.isSealed) {
                                    _context19.next = 8;
                                    break;
                                }

                                throw new Error('The version is not sealed yet');

                            case 8:
                                info = {
                                    'hasEnded': false
                                };

                                //lock

                                _context19.next = 11;
                                return this.redlock.lock('lock:machine:' + machineName + ':version:' + versionKey + ':instanceList', 100);

                            case 11:
                                lock = _context19.sent;
                                _context19.next = 14;
                                return this.client.zcountAsync('machine:' + machineName + ':version:' + versionKey + ':instanceList', '-inf', '+inf');

                            case 14:
                                count = _context19.sent;
                                newInstanceNumber = count + 1;
                                newInstanceName = 'instance' + newInstanceNumber;
                                _context19.next = 19;
                                return this.client.hsetAsync('machine:' + machineName + ':version:' + versionKey + ':instance:' + newInstanceName, 'info', _JSON$stringify(info));

                            case 19:
                                _context19.next = 21;
                                return this.client.zaddAsync('machine:' + machineName + ':version:' + versionKey + ':instanceList', newInstanceNumber, newInstanceName);

                            case 21:
                                _context19.next = 23;
                                return lock.unlock();

                            case 23:
                                _context19.next = 25;
                                return this.client.hincrbyAsync('manifest', 'instanceCount', 1);

                            case 25:
                                return _context19.abrupt('return', newInstanceName);

                            case 26:
                            case 'end':
                                return _context19.stop();
                        }
                    }
                }, _callee19, this);
            }));

            function addInstance(_x36, _x37) {
                return _ref19.apply(this, arguments);
            }

            return addInstance;
        }()

        /**
         * Gets the keys of all of the snapshots of the instance of a version of the machine in the repository
         * @method getSnapshotsKeys
         * @memberOf Core
         * @param {String} machineName The name of the machine to get the snapshots's keys
         * @param {String} versionKey The key of the version to get the snapshots's keys
         * @param {String} instanceKey The key of the instance to get the snapshot's keys
         * @returns {Array} An array with all the snapshot's keys of the instance
         */

    }, {
        key: 'getSnapshotsKeys',
        value: function () {
            var _ref20 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee20(machineName, versionKey, instanceKey) {
                return _regeneratorRuntime.wrap(function _callee20$(_context20) {
                    while (1) {
                        switch (_context20.prev = _context20.next) {
                            case 0:
                                _context20.next = 2;
                                return this._verifyInstance(machineName, versionKey, instanceKey);

                            case 2:
                                _context20.next = 4;
                                return this.client.zrangeAsync('machine:' + machineName + ':version:' + versionKey + ':instance:' + instanceKey + ':snapshotList', 0, -1);

                            case 4:
                                return _context20.abrupt('return', _context20.sent);

                            case 5:
                            case 'end':
                                return _context20.stop();
                        }
                    }
                }, _callee20, this);
            }));

            function getSnapshotsKeys(_x38, _x39, _x40) {
                return _ref20.apply(this, arguments);
            }

            return getSnapshotsKeys;
        }()

        /**
         * Retrieve the snapshot's info.json file as a JavasScript Object
         * @method getSnapshotInfo
         * @memberOf Core
         * @param {String} machineName The name of the machine
         * @param {String} versionKey The key of the version
         * @param {String} instanceKey The key of the instance
         * @param {String} snapshotKey The key of the snapshot
         * @returns {Object} The info Object
         */

    }, {
        key: 'getSnapshotInfo',
        value: function () {
            var _ref21 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee21(machineName, versionKey, instanceKey, snapshotKey) {
                var snapshot;
                return _regeneratorRuntime.wrap(function _callee21$(_context21) {
                    while (1) {
                        switch (_context21.prev = _context21.next) {
                            case 0:
                                _context21.next = 2;
                                return this._verifySnapshot(machineName, versionKey, instanceKey, snapshotKey);

                            case 2:
                                _context21.next = 4;
                                return this.client.hgetAsync('machine:' + machineName + ':version:' + versionKey + ':instance:' + instanceKey + ':snapshot:' + snapshotKey, 'info');

                            case 4:
                                snapshot = _context21.sent;
                                return _context21.abrupt('return', JSON.parse(snapshot));

                            case 6:
                            case 'end':
                                return _context21.stop();
                        }
                    }
                }, _callee21, this);
            }));

            function getSnapshotInfo(_x41, _x42, _x43, _x44) {
                return _ref21.apply(this, arguments);
            }

            return getSnapshotInfo;
        }()

        /**
         * Add a new snapshot to an instance of a version of a machine
         * @method addSnapshot
         * @memberOf Core
         * @param {String} machineName The name of the machine
         * @param {String} versionKey The key of the version
         * @param {String} instanceKey The key of the instance
         * @param {Object} info The info Object
         * @returns {Promise} The instance key
         */

    }, {
        key: 'addSnapshot',
        value: function () {
            var _ref22 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee22(machineName, versionKey, instanceKey, info) {
                var lock, count, newSnapshotNumber, newSnapshotName;
                return _regeneratorRuntime.wrap(function _callee22$(_context22) {
                    while (1) {
                        switch (_context22.prev = _context22.next) {
                            case 0:
                                debug('Adding a new snapshot to the ' + instanceKey + ' of the ' + versionKey + ' of the "' + machineName + '" machine');
                                _context22.next = 3;
                                return this._verifyInstance(machineName, versionKey, instanceKey);

                            case 3:
                                _context22.next = 5;
                                return this.redlock.lock('lock:machine:' + machineName + ':version:' + versionKey + ':instance:' + instanceKey + ':snapshotList', 100);

                            case 5:
                                lock = _context22.sent;
                                _context22.next = 8;
                                return this.client.zcountAsync('machine:' + machineName + ':version:' + versionKey + ':instance:' + instanceKey + ':snapshotList', '-inf', '+inf');

                            case 8:
                                count = _context22.sent;
                                newSnapshotNumber = count + 1;
                                newSnapshotName = 'snapshot' + newSnapshotNumber;
                                _context22.next = 13;
                                return this.client.hsetAsync('machine:' + machineName + ':version:' + versionKey + ':instance:' + instanceKey + ':snapshot:' + newSnapshotName, 'info', _JSON$stringify(info));

                            case 13:
                                _context22.next = 15;
                                return this.client.zaddAsync('machine:' + machineName + ':version:' + versionKey + ':instance:' + instanceKey + ':snapshotList', newSnapshotNumber, newSnapshotName);

                            case 15:
                                _context22.next = 17;
                                return lock.unlock();

                            case 17:
                                _context22.next = 19;
                                return this.client.hincrbyAsync('manifest', 'snapshotCount', 1);

                            case 19:
                                return _context22.abrupt('return', newSnapshotName);

                            case 20:
                            case 'end':
                                return _context22.stop();
                        }
                    }
                }, _callee22, this);
            }));

            function addSnapshot(_x45, _x46, _x47, _x48) {
                return _ref22.apply(this, arguments);
            }

            return addSnapshot;
        }()

        // QUEUE JOBS


        //This jobs implies start an instance interpreter in the last known state.
        /**
         * Queues a job to run an instance
         * @param machine The name of the machine
         * @param versionKey The version key
         * @param instanceKey The instance key
         * @param snapshot An optional snapshot to run the instance in a particular state
         * @returns {Promise.<Job>}
         */

    }, {
        key: 'runInstance',
        value: function () {
            var _ref23 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee23(machine, versionKey) {
                var instanceKey = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
                var snapshot = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;
                var instanceKeySent;
                return _regeneratorRuntime.wrap(function _callee23$(_context23) {
                    while (1) {
                        switch (_context23.prev = _context23.next) {
                            case 0:
                                instanceKeySent = instanceKey;

                                if (!(instanceKeySent === null)) {
                                    _context23.next = 5;
                                    break;
                                }

                                _context23.next = 4;
                                return this.addInstance(machine, versionKey);

                            case 4:
                                instanceKeySent = _context23.sent;

                            case 5:
                                _context23.next = 7;
                                return this._verifyInstance(machine, versionKey, instanceKey);

                            case 7:
                                return _context23.abrupt('return', this.queue.create('runInstance', {
                                    machine: machine,
                                    versionKey: versionKey,
                                    instanceKey: instanceKey,
                                    snapshot: snapshot
                                }).priority('high').save());

                            case 8:
                            case 'end':
                                return _context23.stop();
                        }
                    }
                }, _callee23, this);
            }));

            function runInstance(_x49, _x50) {
                return _ref23.apply(this, arguments);
            }

            return runInstance;
        }()
    }, {
        key: 'stopInstance',
        value: function () {
            var _ref24 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee24(machine, versionKey, instanceKey) {
                return _regeneratorRuntime.wrap(function _callee24$(_context24) {
                    while (1) {
                        switch (_context24.prev = _context24.next) {
                            case 0:
                                _context24.next = 2;
                                return this._verifyInstance(machine, versionKey, instanceKey);

                            case 2:
                                return _context24.abrupt('return', this.queue.create('stop:' + machine + ':' + versionKey + ':' + instanceKey, {
                                    machine: machine,
                                    versionKey: versionKey,
                                    instanceKey: instanceKey
                                }).priority('high').save());

                            case 3:
                            case 'end':
                                return _context24.stop();
                        }
                    }
                }, _callee24, this);
            }));

            function stopInstance(_x53, _x54, _x55) {
                return _ref24.apply(this, arguments);
            }

            return stopInstance;
        }()
    }, {
        key: 'sendEvent',
        value: function () {
            var _ref25 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee25(machine, versionKey, instanceKey, event, data) {
                return _regeneratorRuntime.wrap(function _callee25$(_context25) {
                    while (1) {
                        switch (_context25.prev = _context25.next) {
                            case 0:
                                _context25.next = 2;
                                return this._verifyInstance(machine, versionKey, instanceKey);

                            case 2:
                                return _context25.abrupt('return', this.queue.create('event:' + machine + ':' + versionKey + ':' + instanceKey, {
                                    machine: machine,
                                    versionKey: versionKey,
                                    instanceKey: instanceKey,
                                    event: event,
                                    data: data
                                }).priority('high').save());

                            case 3:
                            case 'end':
                                return _context25.stop();
                        }
                    }
                }, _callee25, this);
            }));

            function sendEvent(_x56, _x57, _x58, _x59, _x60) {
                return _ref25.apply(this, arguments);
            }

            return sendEvent;
        }()

        //Processing jobs

    }, {
        key: 'processRunInstance',
        value: function processRunInstance(handler, maxJobs) {
            if (maxJobs) {
                this.queue.process('runInstance', maxJobs, handler);
            } else {
                this.queue.process('runInstance', handler);
            }
        }
    }, {
        key: 'processStopInstance',
        value: function processStopInstance(machineName, versionKey, instanceKey, handler, maxJobs) {
            if (maxJobs) {
                this.queue.process('stop:' + machineName + ':' + versionKey + ':' + instanceKey, maxJobs, handler);
            } else {
                this.queue.process('stop:' + machineName + ':' + versionKey + ':' + instanceKey, handler);
            }
        }
    }, {
        key: 'processSendEvent',
        value: function processSendEvent(machineName, versionKey, instanceKey, handler, maxJobs) {
            if (maxJobs) {
                this.queue.process('event:' + machineName + ':' + versionKey + ':' + instanceKey, maxJobs, handler);
            } else {
                this.queue.process('event:' + machineName + ':' + versionKey + ':' + instanceKey, handler);
            }
        }
    }]);

    return Core;
}();

export default Core;
//# sourceMappingURL=index.es.js.map
